# python3
# -*- coding: utf-8 -*-
# @File    : DatasetSpace.py
# @Desc    : Book 1-1
# @Project : ML-Algorithm
# @Time    : 9/10/19 3:21 PM
# @Author  : Loopy
# @Contact : peter@mail.loopy.tech
# @License : CC BY-NC-SA 4.0 (subject to project license)

import pandas as pd
from tabulate import tabulate


class DatasetSpace:
    """
    Automatically generate the sample_space, hypothesis_space and version_space for a data set
    """

    def __init__(self, dataset):
        """
        :param dataset: pd.DataFrame
        """
        self.__data = dataset
        self.__features = list(self.__data.columns)
        self.__features.remove("target")

    def get_hypothesis_space(self):
        """
        :return: list (list hypothesis (unknown sample_value))
        """
        possible_values = self.get_possible_values()
        for k in possible_values.keys():
            possible_values[k] += ["*"]
        enum = Enumerator([len(possible_values[f]) for f in self.__features])
        hypothesis_space = []

        while True:
            res = [0] * len(self.__features)
            code = enum.next()
            if code == -1:
                break
            for i, e in enumerate(code):
                res[i] = possible_values[self.__features[i]][e]
            hypothesis_space.append(res)
        hypothesis_space.append(["-"] * len(self.__features))
        return hypothesis_space

    def get_sample_space(self):
        """
        :return: list (list sample (unknown sample_value))
        """
        possible_values = self.get_possible_values()
        enum = Enumerator([len(possible_values[f]) for f in self.__features])
        sample_space = []

        while True:
            res = [0] * len(self.__features)
            code = enum.next()
            if code == -1:
                break
            for i, e in enumerate(code):
                res[i] = possible_values[self.__features[i]][e]
            sample_space.append(res)
        return sample_space

    def get_version_space(self):
        """
        Check the hypothesis_space with the samples, which will get version_space
        :return:
        """
        global hypothesis_space, is_vaild
        hypothesis_space = self.get_hypothesis_space()
        is_vaild = [True] * len(hypothesis_space)

        def check(sample):
            global hypothesis_space, is_vaild
            if sample["target"]:  # positive sample
                for i, hps in enumerate(hypothesis_space):
                    for j, f in enumerate(self.__features):
                        if sample[f] != hps[j] and hps[j] != "*":
                            # hypothesis not agree with the positive sample
                            is_vaild[i] = False
                            break
            else:
                for i, hps in enumerate(hypothesis_space):
                    for j, f in enumerate(self.__features):
                        if sample[f] == hps[j]:
                            # hypothesis agree with the negative sample
                            is_vaild[i] = False
                            break
                    if set(hps) - set(["*"] * len(self.__features)) == set():
                        is_vaild[i] = False
            return sample

        self.__data.apply(check, axis=1)
        version_space = []
        for i, flag in enumerate(is_vaild):
            if flag:
                version_space.append(hypothesis_space[i])
        return version_space

    def get_possible_values(self):
        """
        Get the possible sample values
        :return: dic (key:(str feature)
                      value:(list possible_feature_values))
        """
        possible_values = {}
        for f in self.__features:
            possible_values[f] = list(self.__data[f].unique())
        return possible_values

    @property
    def hypothesis(self):
        return tabulate(
            pd.DataFrame(self.get_hypothesis_space(), columns=self.__features),
            tablefmt="pipe",
            headers="keys",
        )

    @property
    def version(self):
        return tabulate(
            pd.DataFrame(self.get_version_space(), columns=self.__features),
            tablefmt="pipe",
            headers="keys",
        )

    @property
    def sample(self):
        return tabulate(
            pd.DataFrame(self.get_sample_space(), columns=self.__features),
            tablefmt="pipe",
            headers="keys",
        )


class Enumerator:
    """
    enumerator for hypothesis_space
    """

    def __init__(self, max):
        self.code = [0] * len(max)
        self.base = max
        self.overflow = False

    def next(self):
        if self.overflow:
            return -1

        self.code[0] += 1
        if not self.carry():
            self.overflow = True
            return self.code
        else:
            return self.code

    def carry(self):
        for i, b in enumerate(self.base):
            if self.code[i] == b:
                self.code[i] = 0
                if i + 1 >= len(self.code):
                    return False
                self.code[i + 1] += 1

                self.carry()
            else:
                return True


if __name__ == "__main__":
    melon_data = pd.DataFrame(
        [
            ["青绿", "蜷缩", "浊响", True],
            ["乌黑", "蜷缩", "浊响", True],
            ["青绿", "硬挺", "清脆", False],
            ["乌黑", "稍蜷", "沉闷", False],
        ],
        columns=["色泽", "根蒂", "敲声", "target"],
    )

    space = DatasetSpace(melon_data)

    print("=" * 79 + "\n样本空间: \n" + space.sample)

    print("=" * 79 + "\n合取假设的假设空间: \n" + space.hypothesis)

    print("=" * 79 + "\n合取假设的版本空间: \n" + space.version)



import numpy as np
from scipy.optimize import minimize


class LinearRegression:
    """ Linear Regressor
    Dependency
    ----------
    scipy :
        - scipy.optimize.minimize
    numpy:
        - numpy.dot
        - numpy.mean
        - numpy.unique
        - numpy.zeros

    Parameters
    ----------
    None
    Methods
    --------
    fit(self, X, y) : Fit the model using X as training data and y as target values
    predict(self, X) : Predict the class labels for the provided data
    """

    @staticmethod
    def cost_gradient(theta, X, y):
        h = np.dot(X, theta) - y
        J = np.dot(h, h) / (2 * X.shape[0])
        grad = np.dot(X.T, h) / X.shape[0]
        return J, grad

    def fit(self, X, y):
        self._class, y = np.unique(y, return_inverse=True)
        y = y * 2 - 1  # map to(-1,1)
        X_mean, y_mean = np.mean(X, axis=0), np.mean(y)
        X_train, y_train = X - X_mean, y - y_mean

        res = minimize(
            fun=LogisticRegression.cost_gradient,
            jac=True,
            x0=np.zeros(X_train.shape[1]),
            args=(X_train, y_train),
            method="L-BFGS-B",
        )

        self.coef_ = res.x
        self.intercept_ = y_mean - np.dot(X_mean, self.coef_)
        return self

    def predict(self, X):
        y_pred = (
            (1 / (1 + np.exp(-np.dot(X, self.coef_) - self.intercept_))) > 0.5
        ).astype(int)
        return self._class[y_pred]



from .DecisionTreeClassifier import DecisionTreeClassifier
from .DecisionTreeRegressor import DecisionTreeRegressor
from .DatasetSpace import DatasetSpace
from .KNeighborsClassifier import KNeighborsClassifier
from .LinearDiscriminantAnalysis import LinearDiscriminantAnalysis
from .LogisticRegression import LogisticRegression
from .LinearRegression import LinearRegression
from .UnionHypothesisSpace import UnionHypothesisSpace
from .Tao import TAO
from .MelonData import load_melon



import numpy as np


class LinearDiscriminantAnalysis:
    """ Linear Discriminant Analysis
    Dependency
    ----------
    numpy:
        - numpy.dot
        - numpy.mean
        - numpy.unique
        - numpy.linalg.inv

    Parameters
    ----------
    None
    Methods
    --------
    fit(self, X, y) : Fit the model using X as training data and y as target values
    predict(self, X) : Predict the class labels for the provided data
    """

    def fit(self, X, y):
        self._class, y = np.unique(y, return_inverse=True)
        X_p, X_n = X[y == 1], X[y == 0]
        X_p_mean, X_n_mean = X_p.mean(), X_n.mean()

        # 类内散度
        sw = np.array(
            (X_p - X_p_mean).T.dot(X_p - X_p_mean)
            + (X_n - X_n_mean).T.dot(X_n - X_n_mean)
        )

        # 权重判别
        self.w_ = np.dot((X_n_mean - X_p_mean), (np.linalg.inv(sw)))

        # 样本集中心
        self.X_mean_ = (
            np.dot((np.dot(X_p_mean - X_n_mean, np.mat(sw).I)), (X_p_mean + X_n_mean))
            * 0.5
        )

        return self

    def predict(self, X):
        y_pred = ((np.dot(X, self.w_) + self.X_mean_) < 0).astype(int).T
        return self._class[y_pred]



# python3
# -*- coding: utf-8 -*-
# @File    : DecisionTreeClassifier.py
# @Desc    :
# @Project : ML-Algorithm
# @Time    : 11/8/19 10:48 PM
# @Author  : Loopy
# @Contact : peter@mail.loopy.tech
# @License : CC BY-NC-SA 4.0 (subject to project license)

import numpy as np


class TreeNode:
    def __init__(
        self, feature=None, threshold=None, impurity=None, n_node=None, value=None
    ):
        self.left_child = -1
        self.right_child = -1
        self.feature = feature
        self.threshold = threshold
        self.impurity = impurity
        self.n_node = n_node
        self.value = value


class DecisionTreeClassifier:
    def __init__(self, max_depth=2):
        self.max_depth = max_depth

    def _entropy(self, y_cnt):
        prob = y_cnt / np.sum(y_cnt)
        prob = prob[prob > 0]
        return -np.sum(prob * np.log(prob))

    @staticmethod
    def _inti(X, f):
        _, counts = np.unique(X[:, f], return_counts=True)
        ave = counts / counts.sum()
        return (np.log(ave) * ave).sum()

    def _build_leaf(self, X, y, cur_depth, parent, is_left):
        value = np.bincount(y, minlength=self.n_classes_)
        self._nodes.append(TreeNode(n_node=X.shape[0], value=value, impurity=value))
        self._set_parent(parent, is_left, len(self._nodes) - 1)
        return

    def _set_parent(self, parent, is_left, child_ind):
        if parent is not None:
            if is_left:
                self._nodes[parent].left_child = child_ind
            else:
                self._nodes[parent].right_child = child_ind
        return

    def _build_tree(self, X, y, cur_depth, parent, is_left):
        if cur_depth == self.max_depth:
            self._build_leaf(X, y, cur_depth, parent, is_left)
            return

        best_improvement = -np.inf
        best_feature = None
        best_threshold = None
        best_left_ind = None
        best_right_ind = None

        y_cnt = np.bincount(y, minlength=self.n_classes_)
        for i in range(X.shape[1]):
            inti = self._inti(X, i)
            ind = np.argsort(X[:, i])
            y_cnt_left, y_cnt_right = (
                np.bincount([], minlength=self.n_classes_),
                y_cnt.copy(),
            )
            n_left, n_right = 0, X.shape[0]

            for j in range(ind.shape[0] - 1):
                y_cnt_left[y[ind[j]]] += 1
                y_cnt_right[y[ind[j]]] -= 1
                n_left += 1
                n_right -= 1
                if j + 1 < ind.shape[0] - 1 and np.isclose(
                    X[ind[j], i], X[ind[j + 1], i]
                ):
                    continue
                cur_improvement = (
                    n_left * self._entropy(y_cnt_left)
                    + n_right * self._entropy(y_cnt_right)
                ) / inti

                if cur_improvement > best_improvement:
                    best_improvement = cur_improvement
                    best_feature = i
                    best_threshold = X[ind[j], i]
                    best_left_ind = ind[: j + 1]
                    best_right_ind = ind[j + 1 :]

        self._nodes.append(
            TreeNode(
                feature=best_feature,
                threshold=best_threshold,
                n_node=X.shape[0],
                value=y_cnt,
                impurity=self._entropy(y_cnt),
            )
        )

        self._set_parent(parent, is_left, len(self._nodes) - 1)
        if cur_depth < self.max_depth:
            self._build_tree(
                X[best_left_ind],
                y[best_left_ind],
                cur_depth + 1,
                len(self._nodes) - 1,
                True,
            )
            self._build_tree(
                X[best_right_ind],
                y[best_right_ind],
                cur_depth + 1,
                len(self._nodes) - 1,
                False,
            )

    def fit(self, X, y):
        self.n_features = X.shape[1]
        self.classes_, y_train = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        self._nodes = []
        self._build_tree(X, y_train, 0, None, None)
        return self

    def predict(self, X):
        pred = np.zeros(X.shape[0], dtype=np.int)
        for i in range(X.shape[0]):
            cur_node = 0
            while self._nodes[cur_node].left_child != -1:
                if (
                    X[i][self._nodes[cur_node].feature]
                    <= self._nodes[cur_node].threshold
                ):
                    cur_node = self._nodes[cur_node].left_child
                else:
                    cur_node = self._nodes[cur_node].right_child
            pred[i] = cur_node
        return np.array([self.classes_[np.argmax(self._nodes[p].value)] for p in pred])


from sklearn.datasets import load_breast_cancer
from sklearn.tree import DecisionTreeClassifier as skDecisionTreeClassifier

if __name__ == "__main__":
    X, y = load_breast_cancer(return_X_y=True)
    clf1 = DecisionTreeClassifier(max_depth=1).fit(X, y)
    clf2 = skDecisionTreeClassifier(max_depth=1, random_state=0).fit(X, y)

    pred1 = clf1.predict(X)
    pred2 = clf2.predict(X)
    print(pred1)



import scipy
from sklearn.neighbors import BallTree, KDTree


class KNeighborsClassifier:
    """ 一个仿scikit-learn接口的微型knn实现。
    Ｄependency
    ----------
    `scipy` : 考虑到性能，使用scipy来完成底层运算
        - `scipy.unique`
        - `scipy.spatial.distance.cdist`
        - `scipy.argsort`
        - `scipy.stats.mode`
    `sklearn.neighbors` ： 由于时间有限暂时还没有实现的数据结构(TODO: implement of BallTree & KDTree)
        - `sklearn.neighbors.BallTree`
        - `sklearn.neighbors.KDTree`
    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.
    algorithm : {'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:
        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDTree`
        - 'brute' will use a brute-force search.
        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.
    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.
    Methods
    --------
    __init__(n_neighbors=5, algorithm=’brute’, leaf_size=30) : Init
    fit(self.X,y) : Fit the model using X as training data and y as target values
    predict(self, X) : Predict the class labels for the provided data
    Examples
    --------
    >>> X = [[0], [1], [2], [3]]
    >>> y = [0, 0, 1, 1]
    >>> neigh = KNeighborsClassifier(n_neighbors=3)
    >>> neigh.fit(X, y)
    KNeighborsClassifier(...)
    >>> print(neigh.predict([[1.1]]))
    [0]
    >>> print(neigh.predict_proba([[0.9]]))
    [[0.66666667 0.33333333]]
    """

    class KNeighborsClassifier:
        def __init__(self, n_neighbors=5, algorithm="brute", leaf_size=30):
            self.n_neighbors = n_neighbors
            self.algorithm = algorithm
            self.leaf_size = leaf_size

        def fit(self, X, y):
            self._fit_X = X
            self.classes_, self._fit_y = scipy.unique(y, return_inverse=True)

            if self.algorithm == "brute":
                pass
            elif self.algorithm == "kd_tree":
                self.tree = KDTree(X, leaf_size=self.leaf_size)
            elif self.algorithm == "ball_tree":
                self.tree = BallTree(X, leaf_size=self.leaf_size)
            else:
                raise ValueError("unrecognized algorithm: ", str(self.algorithm))
            return self

        def predict(self, X):
            if self.algorithm == "brute":
                dist_mat = scipy.spatial.distance.cdist(X, self._fit_X)
                neighbors_ind = scipy.argsort(dist_mat, axis=1)[:, : self.n_neighbors]
            elif self.algorithm in ["kd_tree", "ball_tree"]:
                dist_mat, neighbors_ind = self.tree.query(X, k=self.n_neighbors)
            else:
                raise ValueError("unrecognized algorithm: ", str(self.algorithm))
            neighbors_y = scipy.stats.mode(self._fit_y[neighbors_ind], axis=1)[
                0
            ].flatten()
            return self.classes_[neighbors_y]



import numpy as np


class TreeNode:
    def __init__(
        self, feature=None, threshold=None, impurity=None, n_node=None, value=None
    ):
        self.left_child = self.right_child = -1
        self.feature = feature
        self.threshold = threshold
        self.impurity = impurity
        self.n_node = n_node
        self.value = value


class DecisionTreeRegressor:
    def __init__(self, max_depth=2):
        self.max_depth = max_depth
        self._nodes = []

    def _build_leaf(self, X, y, cur_depth, parent, is_left):
        self._nodes.append(
            TreeNode(
                impurity=np.mean(np.square(y)) - np.square(np.mean(y)),
                n_node=X.shape[0],
                value=np.mean(y),
            )
        )
        self._set_parent(parent, is_left, len(self._nodes) - 1)
        return

    def _set_parent(self, parent, is_left, child_ind):
        if parent is not None:
            if is_left:
                self._nodes[parent].left_child = len(self._nodes) - 1
            else:
                self._nodes[parent].right_child = len(self._nodes) - 1
        return

    def _build_tree(self, X, y, cur_depth, parent, is_left):
        if cur_depth == self.max_depth:
            self._build_leaf(X, y, cur_depth, parent, is_left)

        best_gain = -np.inf
        best_feature = None
        best_threshold = None
        best_left_ind = best_right_ind = None

        sum_all = np.sum(y)
        step = lambda x, y, a: (x + a, y - a)

        for i in range(X.shape[1]):  # for features
            sum_left, sum_right = 0, sum_all
            n_left = 0
            n_right = X.shape[0]
            ind = np.argsort(X[:, i])

            for j in range(ind.shape[0] - 1):  # for all sample
                # step by step
                sum_left, sum_right = step(sum_left, sum_right, y[ind[j]])
                n_left, n_right = step(n_left, n_right, 1)

                cur_gain = (
                    sum_left * sum_left / n_left + sum_right * sum_right / n_right
                )
                if cur_gain > best_gain:  # found better choice
                    best_gain = cur_gain
                    best_feature = i
                    best_threshold = X[ind[j], i]
                    best_left_ind, best_right_ind = ind[: j + 1], ind[j + 1 :]

        self._nodes.append(
            TreeNode(
                feature=best_feature,
                threshold=best_threshold,
                impurity=np.mean(np.square(y)) - np.square(np.mean(y)),
                n_node=X.shape[0],
                value=np.mean(y),
            )
        )
        cur_id = len(self._nodes) - 1
        self._set_parent(parent, is_left, cur_id)

        if cur_depth < self.max_depth:
            self._build_tree(
                X[best_left_ind], y[best_left_ind], cur_depth + 1, cur_id, True
            )
            self._build_tree(
                X[best_right_ind], y[best_right_ind], cur_depth + 1, cur_id, False
            )

    def fit(self, X, y):
        self.n_features = X.shape[1]
        self._nodes = []
        self._build_tree(X, y, 0, None, None)
        return self

    def predict(self, X):
        pred = np.zeros(X.shape[0])
        for i in range(X.shape[0]):
            cur_node = 0
            while (
                self._nodes[cur_node].left_child != -1
                and self._nodes[cur_node].threshold is not None
            ):  # search in tree
                if (
                    X[i][self._nodes[cur_node].feature]
                    <= self._nodes[cur_node].threshold
                ):
                    cur_node = self._nodes[cur_node].left_child
                else:
                    cur_node = self._nodes[cur_node].right_child
            pred[i] = self._nodes[cur_node].value
        return pred



# A special gift for Professor Hu

from requests import get
import multiprocessing
import time
import datetime
from tqdm import tqdm


class TAO:
    NUM_OF_PROCESS = 100
    INTERVAL = 120

    @staticmethod
    def click_hzs():
        while "天下难事，必作于易；天下大事，必作于细":
            get(
                "http://www.cs.cqu.edu.cn/system/resource/code/datainput.jsp?owner=1331212424&newsid=4215"
            )
        return "大方无隅，大器晚成，大音希声，大象无形"

    @staticmethod
    def get_hzs():
        while "祸兮，福之所倚；福兮，祸之所伏":
            r = get(
                "http://www.cs.cqu.edu.cn/system/resource/code/news/click/dynclicks.jsp?clickid=4212&owner=1331212424&clicktype=wbnews"
            )

            if r.status_code == 200:
                return int(r.content.decode("utf-8")) / 1000
        return "道，可道，非常道；名，可名，非常名"

    def __init__(self, num_of_process=None, interval=None):
        print(
            """
        TTTTTTTTTTTTTTTTTTTTTTT                                
        T:::::::::::::::::::::T                                
        T:::::::::::::::::::::T                                
        T:::::TT:::::::TT:::::T                                
        TTTTTT  T:::::T  TTTTTTaaaaaaaaaaaaa     ooooooooooo   
                T:::::T        a::::::::::::a  oo:::::::::::oo 
                T:::::T        aaaaaaaaa:::::ao:::::::::::::::o
                T:::::T                 a::::ao:::::ooooo:::::o
                T:::::T          aaaaaaa:::::ao::::o     o::::o
                T:::::T        aa::::::::::::ao::::o     o::::o
                T:::::T       a::::aaaa::::::ao::::o     o::::o
                T:::::T      a::::a    a:::::ao::::o     o::::o
              TT:::::::TT    a::::a    a:::::ao:::::ooooo:::::o
              T:::::::::T    a:::::aaaa::::::ao:::::::::::::::o
              T:::::::::T     a::::::::::aa:::aoo:::::::::::oo 
              TTTTTTTTTTT      aaaaaaaaaa  aaaa  ooooooooooo   \n\n"""
        )

        TAO.NUM_OF_PROCESS = (
            num_of_process if num_of_process is not None else TAO.NUM_OF_PROCESS
        )
        TAO.INTERVAL = interval if interval is not None else TAO.INTERVAL

        # create process_pool
        process_pool = []

        print(
            "\033[1;31;40m[1] Tao gave birth to the One\033[0m Create process_pool, number of process:{}".format(
                TAO.NUM_OF_PROCESS
            )
        )
        for i in tqdm(range(TAO.NUM_OF_PROCESS)):
            process_pool.append(multiprocessing.Process(target=self.click_hzs, args=()))

        # activate process in process_pool
        print(
            "\033[1;31;40m[2] One gave birth successively to two\033[0m Activate process in process_pool"
        )
        for i in tqdm(range(TAO.NUM_OF_PROCESS)):
            process_pool[i].start()

        # get the result
        print(
            "\033[1;31;40m[3] Two gave birth successively to three, then infinity arise\033[0m\n"
            "Congratulations! Tao is running now!! \n"
            "Follows are what we have achieved (The result will refresh every {} seconds):\n".format(
                TAO.INTERVAL
            )
            + "=" * 77
        )
        res = self.get_hzs()
        print(
            "[{}] So far it has had {}k hits".format(
                datetime.datetime.now().strftime("%y-%m-%d %I:%M:%S %p"), round(res, 2),
            )
        )
        while True:
            pre_time = datetime.datetime.now()
            pre_res = res

            time.sleep(TAO.INTERVAL)

            res = self.get_hzs()
            print(
                "[{}] So far it has had {}k hits, which are growing at {}k hits/s".format(
                    datetime.datetime.now().strftime("%y-%m-%d %I:%M:%S %p"),
                    round(res, 2),
                    round(
                        (res - pre_res) / (datetime.datetime.now() - pre_time).seconds,
                        4,
                    ),
                )
            )



# python3
# -*- coding: utf-8 -*-
# @File    : UnionHypothesisSpace.py
# @Desc    : Book 1-2
# @Project : ML-Algorithm
# @Time    : 9/10/19 6:05 PM
# @Author  : Loopy
# @Contact : peter@mail.loopy.tech
# @License : CC BY-NC-SA 4.0 (subject to project license)

from .DatasetSpace import DatasetSpace
import json
import pandas as pd
import matplotlib.pyplot as plt


class UnionHypothesisSpace:
    """
    Figure out how the number of hypothsis change with the disjunction become longer
    """

    def __init__(self, dataset=None):
        if dataset is None:
            return
        self.__data = dataset
        self.__space = DatasetSpace(dataset)

        self.__features = list(self.__data.columns)
        self.__features.remove("target")

        self.sample_space = self.get_sample_space()

        # history hypothsis_code
        self.his_hypothsis_code = dict()
        self.his_hypothsis_code["k=1"] = self.get_conj_hypothesis()
        for i, h in enumerate(self.his_hypothsis_code["k=1"]):
            self.his_hypothsis_code["k=1"][i] = self.encode(h)

        # shown hypothsis_codes
        self.hypothsis_code_pool = [False] * 2 ** (len(self.sample_space))

    def get_conj_hypothesis(self):
        return self.__space.get_hypothesis_space()

    def get_sample_space(self):
        return self.__space.get_sample_space()

    def encode(self, hypothesis):
        """
        Encode a hypothsis to a unique code
        :param hypothesis: list
        :return: int the unique code
        """
        code = ""
        for s in self.sample_space:
            flag = "1"
            for i in range(len(self.__features)):
                if s[i] != hypothesis[i] and hypothesis[i] != "*":
                    flag = "0"
            code += flag
        return int(code, 2)

    def union(self, k):
        """
        Try union the hypothsis and skip redundant ones
        :param k:
        :return:
        """
        self.his_hypothsis_code["k=" + str(k + 1)] = []
        for i, h_l in enumerate(self.his_hypothsis_code["k=" + str(k)]):
            for j, h_r in enumerate(self.his_hypothsis_code["k=1"]):
                if not self.hypothsis_code_pool[(h_l | h_r)]:
                    self.hypothsis_code_pool[(h_l | h_r)] = True
                    self.his_hypothsis_code["k=" + str(k + 1)].append((h_l | h_r))
                print("k={} i={} j={}".format(k, i, j), end="\r")

    def run(self, k=1):
        """
        runner
        :param k: int the k to start
        :return: None
        """
        max_hypothsis_number = 2 ** (len(self.sample_space))
        count = 0
        while count < max_hypothsis_number:
            self.union(k)
            self.save_report(k)
            k += 1

            count = 0
            for h in self.hypothsis_code_pool:
                if h:
                    count += 1
        self.plot(k)

    def load(self):
        """
        load result from res.json
        :return:
        """
        with open("../temp/res.json", "r") as f:
            res = json.loads(f.read())

        self.__data = res["dataset"]
        self.__space = DatasetSpace(res["dataset"])
        self.his_hypothsis_code = res["his_hypothsis_code"]
        self.hypothsis_code_pool = res["hypothsis_code_pool"]

        self.sample_space = self.get_sample_space()

        self.__features = list(self.__data.columns)
        self.__features.remove("target")

    def save_report(self, k):
        """
        save result to res.json and report
        :param k: int current k
        :return:  None
        """
        with open("../temp/res.json", "w") as f:
            res = str(
                {
                    "dataset": self.__data,
                    "his_hypothsis_code": self.his_hypothsis_code,
                    "hypothsis_code_pool": self.hypothsis_code_pool,
                }
            )

            f.write(res)
        count = 0
        for h in self.hypothsis_code_pool:
            if h:
                count += 1
        print(
            "{}\n"
            "Working on {}-term-disjunction:\n"
            "Number of {}-term-disjunction hypothsis:{}\n"
            "Number of hypothsis:{}/{} = {}%".format(
                "=" * 79,
                k + 1,
                k,
                len(self.his_hypothsis_code["k=" + str(k)]),
                count,
                2 ** (len(self.sample_space)),
                round(count / 2 ** (len(self.sample_space)) * 100, 2),
            )
        )

    def plot(self, k):
        """
        plot k-number of hypothsis
        :param k: int max number of disjunction term
        :return: None
        """
        x = list(range(k))
        y = [len(self.his_hypothsis_code["k=" + str(i + 1)]) for i in range(k)]
        plt.plot(x, y)
        plt.show()
        plt.savefig("../temp/1.png")


if __name__ == "__main__":
    melon_data = pd.DataFrame(
        [
            ["青绿", "蜷缩", "浊响", True],
            ["乌黑", "蜷缩", "浊响", True],
            ["青绿", "硬挺", "清脆", False],
            ["乌黑", "稍蜷", "沉闷", False],
        ],
        columns=["色泽", "根蒂", "敲声", "target"],
    )

    space = UnionHypothesisSpace(melon_data)
    space.run()

    # when you have to continue
    # space = UnionHypothesisSpace()
    # space.load()
    # space.run(10)



#!/usr/bin/env python3
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='cquai-ml',
    version='1.0.6',
    description=(
        'Implement machine learning algorithms with python without sklearn.'
    ),
    long_description=open('README.rst').read(),
    author='CQU-AI',
    author_email='peter@mail.loopy.tech',
    maintainer='loopyme',
    maintainer_email='peter@mail.loopy.tech',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/CQU-AI/ML-Algorithm',
    install_requires=[
        'tabulate>=0.8.5',
        'tqdm>=4.38.0',
        'requests>=2.22.0',
        'numpy>=1.17.4',
        'pandas>=0.25.3',
        'sklearn>=0.0'
    ]
)



import numpy as np
from scipy.optimize import minimize
from scipy.special import expit


class LogisticRegression:
    """ Binary Logistic regression based on gradient descent (l-bfgs-b) without penalty
    Dependency
    ----------
    scipy :
        - scipy.optimize.minimize
        - scipy.special.expit
    numpy:
        - numpy.dot
        - numpy.unique
        - numpy.zeros

    Parameters
    ----------
    None
    Methods
    --------
    fit(self, X, y) : Fit the model using X as training data and y as target values
    predict(self, X) : Predict the class labels for the provided data
    """

    @staticmethod
    def cost_gradient(w, X, y):
        yz = y * (np.dot(X, w[:-1]) + w[-1])
        cost = -np.sum(
            np.vectorize(
                lambda x: -np.log(1 + np.exp(-x))
                if x > 0
                else x - np.log(1 + np.exp(x))
            )(yz)
        ) + 0.5 * np.dot(w[:-1], w[:-1])
        grad = np.zeros(len(w))
        t = (expit(yz) - 1) * y
        grad[:-1] = np.dot(X.T, t) + w[:-1]
        grad[-1] = np.sum(t)
        return cost, grad

    def fit(self, X, y):
        self._class, y = np.unique(y, return_inverse=True)
        y = y * 2 - 1  # map to(-1,1)

        res = minimize(
            fun=LogisticRegression.cost_gradient,
            jac=True,
            x0=np.zeros(X.shape[1] + 1),
            args=(X, y),
            method="L-BFGS-B",
        )
        self.coef_, self.intercept_ = res.x[:-1], res.x[-1]
        return self

    def predict(self, X):
        y_pred = ((np.dot(X, self.coef_) + self.intercept_) > 0).astype(int)
        return self._class[y_pred]
